package sun.reflect.misc;

import java.security.AllPermission;
import java.security.PermissionCollection;
import java.net.URLConnection;
import sun.misc.IOUtils;
import java.io.BufferedInputStream;
import java.net.HttpURLConnection;
import java.security.CodeSource;
import java.security.cert.Certificate;
import java.net.URL;
import java.io.IOException;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Modifier;
import java.util.Map;
import java.util.HashMap;
import java.lang.reflect.Method;
import java.security.SecureClassLoader;

public final class MethodUtil extends SecureClassLoader
{
    private static final String MISC_PKG = "sun.reflect.misc.";
    private static final String TRAMPOLINE = "sun.reflect.misc.Trampoline";
    private static final Method bounce;
    
    private MethodUtil() {
    }
    
    public static Method getMethod(Class<?> clazz, String name, Class<?>[] parameterTypes) throws NoSuchMethodException {
        ReflectUtil.checkPackageAccess(clazz);
        return clazz.getMethod(name, parameterTypes);
    }
    
    public static Method[] getMethods(Class<?> clazz) {
        ReflectUtil.checkPackageAccess(clazz);
        return clazz.getMethods();
    }
    
    public static Method[] getPublicMethods(Class<?> superclass) {
        if (System.getSecurityManager() == null) {
            return superclass.getMethods();
        }
        HashMap<MethodUtil.Signature, Method> hashMap;
        for (hashMap = new HashMap(); superclass != null && !getInternalPublicMethods(superclass, hashMap); superclass = superclass.getSuperclass()) {
            getInterfaceMethods(superclass, hashMap);
        }
        return (Method[])hashMap.values().toArray(new Method[hashMap.size()]);
    }
    
    private static void getInterfaceMethods(Class<?> clazz, Map<MethodUtil.Signature, Method> map) {
        Class[] interfaces = clazz.getInterfaces();
        for (int i = 0; i < interfaces.length; ++i) {
            Class<Object> clazz2 = interfaces[i];
            if (!getInternalPublicMethods(clazz2, map)) {
                getInterfaceMethods(clazz2, map);
            }
        }
    }
    
    private static boolean getInternalPublicMethods(Class<?> clazz, Map<MethodUtil.Signature, Method> map) {
        Method[] methods;
        try {
            if (!Modifier.isPublic(clazz.getModifiers())) {
                return false;
            }
            if (!ReflectUtil.isPackageAccessible(clazz)) {
                return false;
            }
            methods = clazz.getMethods();
        }
        catch (SecurityException ex) {
            return false;
        }
        boolean n = true;
        int i = 0;
        while (i < methods.length) {
            if (!Modifier.isPublic(methods[i].getDeclaringClass().getModifiers())) {
                n = false;
                break;
            }
            ++i;
        }
        if (!(n == 0)) {
            for (int k = 0; k < methods.length; ++k) {
                addMethod(map, methods[k]);
            }
        }
        else {
            for (int j = 0; j < methods.length; ++j) {
                if (clazz.equals(methods[j].getDeclaringClass())) {
                    addMethod(map, methods[j]);
                }
            }
        }
        return n;
    }
    
    private static void addMethod(Map<MethodUtil.Signature, Method> map, Method method) {
        MethodUtil.Signature signature = new Signature(method);
        if (!map.containsKey(signature)) {
            map.put(signature, method);
        }
        else if (!method.getDeclaringClass().isInterface() && ((Method)map.get(signature)).getDeclaringClass().isInterface()) {
            map.put(signature, method);
        }
    }
    
    public static Object invoke(Method method, Object o, Object[] array) throws InvocationTargetException, IllegalAccessException {
        try {
            return MethodUtil.bounce.invoke(null, new Object[] { method, o, array });
        }
        catch (InvocationTargetException ex) {
            Throwable cause = ex.getCause();
            if ((cause instanceof InvocationTargetException)) {
                throw (InvocationTargetException)cause;
            }
            if ((cause instanceof IllegalAccessException)) {
                throw (IllegalAccessException)cause;
            }
            if ((cause instanceof RuntimeException)) {
                throw (RuntimeException)cause;
            }
            if ((cause instanceof Error)) {
                throw (Error)cause;
            }
            throw new Error("Unexpected invocation error", cause);
        }
        catch (IllegalAccessException cause2) {
            throw new Error("Unexpected invocation error", cause2);
        }
    }
    
    private static Method getTrampoline() {
        try {
            return (Method)AccessController.doPrivileged(new PrivilegedExceptionAction<Method>() {
                @Override
                public Method run() throws Exception {
                    Method declaredMethod = MethodUtil.access$000().getDeclaredMethod("invoke", new Class[] { Method.class, Object.class, Object[].class });
                    declaredMethod.setAccessible(true);
                    return declaredMethod;
                }
            });
        }
        catch (Exception cause) {
            throw new InternalError("bouncer cannot be found", cause);
        }
    }
    
    @Override
    protected synchronized Class<?> loadClass(String s, boolean b) throws ClassNotFoundException {
        ReflectUtil.checkPackageAccess(s);
        Class<Object> c = this.findLoadedClass(s);
        if (c == null) {
            try {
                c = this.findClass(s);
            }
            catch (ClassNotFoundException ex) {}
            if (c == null) {
                c = this.getParent().loadClass(s);
            }
        }
        if (b) {
            this.resolveClass(c);
            return c;
        }
        return c;
    }
    
    @Override
    protected Class<?> findClass(String s) throws ClassNotFoundException {
        if (!s.startsWith("sun.reflect.misc.")) {
            throw new ClassNotFoundException(s);
        }
        URL resource = this.getResource(s.replace('.', '/').concat(".class"));
        if (resource != null) {
            try {
                return this.defineClass(s, resource);
            }
            catch (IOException ex) {
                throw new ClassNotFoundException(s, ex);
            }
        }
        else {
            throw new ClassNotFoundException(s);
        }
    }
    
    private Class<?> defineClass(String s, URL url) throws IOException {
        byte[] bytes = getBytes(url);
        CodeSource cs = new CodeSource(null, (Certificate[])null);
        if (!s.equals("sun.reflect.misc.Trampoline")) {
            throw new IOException("MethodUtil: bad name " + s);
        }
        return super.defineClass(s, bytes, 0, bytes.length, cs);
    }
    
    private static byte[] getBytes(URL url) throws IOException {
        URLConnection openConnection = url.openConnection();
        if (openConnection instanceof HttpURLConnection && ((HttpURLConnection)openConnection).getResponseCode() >= 400) {
            throw new IOException("open HTTP connection failed.");
        }
        int contentLength = openConnection.getContentLength();
        BufferedInputStream o = new BufferedInputStream(openConnection.getInputStream());
        byte[] fully;
        try {
            fully = IOUtils.readFully(o, contentLength, true);
        }
        finally {
            o.close();
        }
        return fully;
    }
    
    @Override
    protected PermissionCollection getPermissions(CodeSource codesource) {
        PermissionCollection permissions = super.getPermissions(codesource);
        permissions.add(new AllPermission());
        return permissions;
    }
    
    private static Class<?> getTrampolineClass() {
        try {
            return Class.forName("sun.reflect.misc.Trampoline", true, new MethodUtil());
        }
        catch (ClassNotFoundException ex) {
            return null;
        }
    }
    
    static Class access$000() {
        return getTrampolineClass();
    }
    
    static {
        bounce = getTrampoline();
    }
    
    private static class Signature
    {
        private String methodName;
        private Class<?>[] argClasses;
        private volatile int hashCode;
        
        Signature(Method method) {
            this.hashCode = 0;
            this.methodName = method.getName();
            this.argClasses = method.getParameterTypes();
        }
        
        @Override
        public boolean equals(Object o) {
            if (this == o) {
                return true;
            }
            MethodUtil.Signature signature = (MethodUtil.Signature)o;
            if (!this.methodName.equals(signature.methodName)) {
                return false;
            }
            if (this.argClasses.length != signature.argClasses.length) {
                return false;
            }
            int i = 0;
            while (i < this.argClasses.length) {
                if (this.argClasses[i] != signature.argClasses[i]) {
                    return false;
                }
                ++i;
            }
            return true;
        }
        
        @Override
        public int hashCode() {
            if (this.hashCode == 0) {
                int hashCode = 37 * 17 + this.methodName.hashCode();
                if (this.argClasses != null) {
                    for (int i = 0; i < this.argClasses.length; ++i) {
                        hashCode = 37 * hashCode + ((this.argClasses[i] == null) ? 0 : this.argClasses[i].hashCode());
                    }
                }
                this.hashCode = hashCode;
            }
            return this.hashCode;
        }
    }
}
